#include "async.h"
#include "parseopt.h"

const int fd_max = 256;
// Only accept this many simultaneous clients:
const int proxy_max = (fd_max - 8) / 2;

static int lfd;			// Socket on which we listen for new clients
static sockaddr_in remaddr;	// Address of server we must connect to
static void makeproxy ();

static void
makeproxy ()
{
  sockaddr_in sin;
  socklen_t sinlen = sizeof (sin);
  bzero (&sin, sizeof (sin));
  int cfd = accept (lfd, (sockaddr *) &sin, &sinlen);
  if (cfd >= 0) {
    int sfd = inetsocket (SOCK_STREAM);
    if (sfd >= 0)
      make_async (sfd);
    if (sfd < 0) {
      warn << "socket: " << strerror (errno) << "\n";
      close (cfd);
    }
    else if (connect (sfd, (const sockaddr *) &remaddr, sizeof (remaddr)) < 0
	     && errno != EINPROGRESS) {
      warn << "connect: " << strerror (errno) << "\n";
      close (sfd);
      close (cfd);
    }
    else {
      /* Now you must relay data back and forth between the client (on
       * cfd) and the server (on sfd), instead of just closing the
       * file descriptors. */
      warn << "accepted client from " << inet_ntoa (sin.sin_addr)
	   << " port " << ntohs (sin.sin_port) << "\n";
      close (sfd);		// XXX
      close (cfd);		// XXX
    }
  }
  else if (errno != EAGAIN)
    warn << "accept: " << strerror (errno) << "\n";

  /* If you have more than proxy_max proxies, you should stop
   * accepting new connections and clear the callback on ldf with:
   *
   *    fdcb (lfd, selread, NULL);
   *
   * (but of course you also need to reenable the callback after the
   * number of proxies drops below proxy_max).
   */
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);

  u_int16_t localport, remport;
  if (argc != 4 || !convertint (argv[2], &remport)
      || !convertint (argv[3], &localport)) {
    warnx << "usage: " << progname << " remote-host remote-port local-port\n";
    exit (1);
  }
  hostent *h = gethostbyname (argv[1]);
  if (!h)
    fatal << argv[1] << ": hostname lookup failed\n";
  remaddr.sin_family = AF_INET;
  remaddr.sin_port = htons (remport);
  remaddr.sin_addr = *reinterpret_cast<in_addr *> (h->h_addr);
  
  lfd = inetsocket (SOCK_STREAM, localport);
  if (lfd < 0)
    fatal << "port " << localport << ": " << strerror (errno) << "\n";
  make_async (lfd);
  listen (lfd, 5);
  fdcb (lfd, selread, wrap (&makeproxy));

  sigcb (SIGINT, wrap (exit, 1));
  sigcb (SIGTERM, wrap (exit, 1));

  amain ();
}
